use std::path::{Path, PathBuf};
use std::process::Command;
use std::{env, fs};

use anyhow::{bail, Context, Error};
use console::style;
use same_file::is_same_file;
use serde::{Deserialize, Serialize};
use tempfile::tempdir;

use crate::bootstrap::{ensure_self_venv, fetch, get_pip_module};
use crate::config::Config;
use crate::consts::VENV_BIN;
use crate::lock::{
    make_project_root_fragment, update_single_project_lockfile, update_workspace_lockfile,
    LockMode, LockOptions,
};
use crate::piptools::{get_pip_sync, get_pip_tools_venv_path};
use crate::platform::get_toolchain_python_bin;
use crate::pyproject::{read_venv_marker, ExpandedSources, PyProject};
use crate::sources::PythonVersion;
use crate::utils::{
    get_venv_python_bin, mark_path_sync_ignore, set_proxy_variables, symlink_dir, CommandOutput,
};

/// Controls the sync mode
#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Default)]
pub enum SyncMode {
    /// Just ensures Python is there
    #[default]
    PythonOnly,
    /// Lock only
    LockOnly,
    /// Update dependencies
    Regular,
    /// recreate everything
    Full,
}

/// Updates the virtualenv based on the pyproject.toml
#[derive(Debug, Default)]
pub struct SyncOptions {
    /// How verbose should the sync be?
    pub output: CommandOutput,
    /// Include dev dependencies?
    pub dev: bool,
    /// Which sync mode should be used?
    pub mode: SyncMode,
    /// Forces venv creation even when unsafe.
    pub force: bool,
    /// Do not lock.
    pub no_lock: bool,
    /// Controls locking.
    pub lock_options: LockOptions,
    /// Explicit pyproject location (Only usable by PythonOnly mode)
    pub pyproject: Option<PathBuf>,
}

impl SyncOptions {
    /// Only sync the Python itself.
    pub fn python_only() -> SyncOptions {
        SyncOptions {
            mode: SyncMode::PythonOnly,
            ..Default::default()
        }
    }

    pub fn pyproject(mut self, pyproject: Option<PathBuf>) -> Self {
        self.pyproject = pyproject;
        self
    }
}

/// Config written into the virtualenv for sync purposes.
#[derive(Serialize, Deserialize, Debug)]
pub struct VenvMarker {
    pub python: PythonVersion,
    pub venv_path: Option<PathBuf>,
}

/// Synchronizes a project's virtualenv.
pub fn sync(mut cmd: SyncOptions) -> Result<(), Error> {
    let pyproject = PyProject::load_or_discover(cmd.pyproject.as_deref())?;
    let lockfile = pyproject.workspace_path().join("requirements.lock");
    let dev_lockfile = pyproject.workspace_path().join("requirements-dev.lock");
    let venv = pyproject.venv_path();
    let py_ver = pyproject.venv_python_version()?;
    let output = cmd.output;

    if cmd.pyproject.is_some()
        && cmd.mode != SyncMode::PythonOnly
        && !pyproject.toml_path().ends_with("pyproject.toml")
    {
        // pip-tools will search for pyproject.toml
        bail!("cannot sync or generate lockfile: package needs 'pyproject.toml'");
    }

    // Turn on locking with sources if the project demands it.
    if pyproject.lock_with_sources() {
        cmd.lock_options.with_sources = true;
    }

    // ensure we are bootstrapped
    let self_venv = ensure_self_venv(output).context("could not sync because bootstrap failed")?;

    let mut recreate = cmd.mode == SyncMode::Full;
    if venv.is_dir() {
        if let Some(marker) = read_venv_marker(&venv) {
            if marker.python != py_ver {
                if cmd.output != CommandOutput::Quiet {
                    echo!(
                        "Python version mismatch (found {}, expected {}), recreating.",
                        marker.python,
                        py_ver
                    );
                }
                recreate = true;
            } else if let Some(ref venv_path) = marker.venv_path {
                // for virtualenvs that have a location identifier, check if we need to
                // recreate it.  On IO error we know that one of the paths is gone, so
                // something needs recreation.
                if !is_same_file(&venv, venv_path).unwrap_or(false) {
                    if cmd.output != CommandOutput::Quiet {
                        echo!(
                            "Detected relocated virtualenv ({} => {}), recreating.",
                            venv_path.display(),
                            venv.display(),
                        );
                    }
                    recreate = true;
                }
            }
        } else if cmd.force {
            if cmd.output != CommandOutput::Quiet {
                echo!("Forcing re-creation of non-rye managed virtualenv");
            }
            recreate = true;
        } else if cmd.mode == SyncMode::PythonOnly {
            // in python-only sync mode, don't complain about foreign venvs
            return Ok(());
        } else {
            bail!("virtualenv is not managed by rye. Run `rye sync -f` to force.");
        }
    }

    // make sure we have a compatible python version
    let py_ver =
        fetch(&py_ver.into(), output).context("failed fetching toolchain ahead of sync")?;

    // kill the virtualenv if it's there and we need to get rid of it.
    if recreate && venv.is_dir() {
        fs::remove_dir_all(&venv).context("failed to delete existing virtualenv")?;
    }

    if venv.is_dir() {
        // we only care about this output if regular syncs are used
        if !matches!(cmd.mode, SyncMode::PythonOnly | SyncMode::LockOnly)
            && output != CommandOutput::Quiet
        {
            echo!("Reusing already existing virtualenv");
        }
    } else {
        if output != CommandOutput::Quiet {
            echo!(
                "Initializing new virtualenv in {}",
                style(venv.display()).cyan()
            );
            echo!("Python version: {}", style(&py_ver).cyan());
        }
        let prompt = pyproject.name().unwrap_or("venv");
        create_virtualenv(output, &self_venv, &py_ver, &venv, prompt)
            .context("failed creating virtualenv ahead of sync")?;
        fs::write(
            venv.join("rye-venv.json"),
            serde_json::to_string_pretty(&VenvMarker {
                python: py_ver.clone(),
                venv_path: Some(venv.clone().into()),
            })?,
        )
        .context("failed writing venv marker file")?;
    }

    // prepare necessary utilities for pip-sync.  This is a super crude
    // hack to make this work for now.  We basically sym-link pip itself
    // into a folder all by itself and place a second file in there which we
    // can pass to pip-sync to install the local package.
    if recreate || cmd.mode != SyncMode::PythonOnly {
        let sources = ExpandedSources::from_sources(&pyproject.sources()?)?;
        if cmd.no_lock {
            let lockfile = if cmd.dev { &dev_lockfile } else { &lockfile };
            if !lockfile.is_file() {
                bail!(
                    "Locking is disabled but lockfile '{}' does not exist",
                    lockfile.display()
                );
            }
        } else if let Some(workspace) = pyproject.workspace() {
            // make sure we have an up-to-date lockfile
            update_workspace_lockfile(
                &py_ver,
                workspace,
                LockMode::Production,
                &lockfile,
                cmd.output,
                &sources,
                &cmd.lock_options,
            )
            .context("could not write production lockfile for workspace")?;
            update_workspace_lockfile(
                &py_ver,
                workspace,
                LockMode::Dev,
                &dev_lockfile,
                cmd.output,
                &sources,
                &cmd.lock_options,
            )
            .context("could not write dev lockfile for workspace")?;
        } else {
            // make sure we have an up-to-date lockfile
            update_single_project_lockfile(
                &py_ver,
                &pyproject,
                LockMode::Production,
                &lockfile,
                cmd.output,
                &sources,
                &cmd.lock_options,
            )
            .context("could not write production lockfile for project")?;
            update_single_project_lockfile(
                &py_ver,
                &pyproject,
                LockMode::Dev,
                &dev_lockfile,
                cmd.output,
                &sources,
                &cmd.lock_options,
            )
            .context("could not write dev lockfile for project")?;
        }

        // run pip install with the lockfile.
        if cmd.mode != SyncMode::LockOnly {
            if output != CommandOutput::Quiet {
                echo!("Installing dependencies");
            }

            let tempdir = tempdir()?;
            let mut sync_cmd = if Config::current().use_uv() {
                let mut uv_sync_cmd = Command::new(self_venv.join(VENV_BIN).join("uv"));
                uv_sync_cmd.arg("pip").arg("sync");
                let root = pyproject.workspace_path();

                uv_sync_cmd
                    .env("PROJECT_ROOT", make_project_root_fragment(&root))
                    .env("VIRTUAL_ENV", pyproject.venv_path().as_os_str())
                    .current_dir(&root);
                uv_sync_cmd
            } else {
                let mut pip_sync_cmd = Command::new(get_pip_sync(&py_ver, output)?);
                let root = pyproject.workspace_path();
                let py_path = get_venv_python_bin(&venv);

                // we need to run this after we have run the `get_pip_sync` command
                // as this is what bootstraps or updates the pip tools installation.
                // This is needed as on unix platforms we need to search the module path.
                symlink_dir(
                    get_pip_module(&get_pip_tools_venv_path(&py_ver))
                        .context("could not locate pip")?,
                    tempdir.path().join("pip"),
                )
                .context("failed linking pip module into for pip-sync")?;

                pip_sync_cmd
                    .env("PROJECT_ROOT", make_project_root_fragment(&root))
                    .env("PYTHONPATH", tempdir.path())
                    .current_dir(&root)
                    .arg("--python-executable")
                    .arg(&py_path)
                    .arg("--pip-args")
                    .arg("--no-deps");

                if output != CommandOutput::Quiet {
                    pip_sync_cmd.env("PYTHONWARNINGS", "ignore");
                } else if output == CommandOutput::Verbose && env::var("PIP_VERBOSE").is_err() {
                    pip_sync_cmd.env("PIP_VERBOSE", "2");
                }
                pip_sync_cmd
            };

            sources.add_as_pip_args(&mut sync_cmd);

            if cmd.dev && dev_lockfile.is_file() {
                sync_cmd.arg(&dev_lockfile);
            } else {
                sync_cmd.arg(&lockfile);
            }

            if output == CommandOutput::Verbose {
                sync_cmd.arg("--verbose");
            } else if output == CommandOutput::Quiet {
                sync_cmd.arg("-q");
            }
            set_proxy_variables(&mut sync_cmd);
            let status = sync_cmd.status().context("unable to run pip-sync")?;

            if !status.success() {
                bail!("Installation of dependencies failed");
            }
        }
    }

    if output != CommandOutput::Quiet && cmd.mode != SyncMode::PythonOnly {
        echo!("Done!");
    }

    Ok(())
}

pub fn create_virtualenv(
    output: CommandOutput,
    self_venv: &Path,
    py_ver: &PythonVersion,
    venv: &Path,
    prompt: &str,
) -> Result<(), Error> {
    let py_bin = get_toolchain_python_bin(py_ver)?;

    let mut venv_cmd = if Config::current().use_uv() {
        // try to kill the empty venv if there is one as uv can't work otherwise.
        fs::remove_dir(venv).ok();
        let mut venv_cmd = Command::new(self_venv.join(VENV_BIN).join("uv"));
        venv_cmd.arg("venv");
        if output == CommandOutput::Verbose {
            venv_cmd.arg("--verbose");
        } else {
            venv_cmd.arg("-q");
        }
        venv_cmd.arg("-p");
        venv_cmd.arg(&py_bin);
        venv_cmd
    } else {
        // create the venv folder first so we can manipulate some flags on it.
        fs::create_dir_all(venv)
            .with_context(|| format!("unable to create virtualenv folder '{}'", venv.display()))?;

        update_venv_sync_marker(output, venv);
        let mut venv_cmd = Command::new(self_venv.join(VENV_BIN).join("virtualenv"));
        if output == CommandOutput::Verbose {
            venv_cmd.arg("--verbose");
        } else {
            venv_cmd.arg("-q");
            venv_cmd.env("PYTHONWARNINGS", "ignore");
        }
        venv_cmd.arg("-p");
        venv_cmd.arg(&py_bin);
        venv_cmd.arg("--no-seed");
        venv_cmd.arg("--prompt");
        venv_cmd.arg(prompt);
        venv_cmd
    };

    venv_cmd.arg("--").arg(venv);

    let status = venv_cmd
        .status()
        .context("unable to invoke virtualenv command")?;
    if !status.success() {
        bail!("failed to initialize virtualenv");
    }

    // uv can only do it now
    if Config::current().use_uv() {
        update_venv_sync_marker(output, venv);
    }

    // On UNIX systems Python is unable to find the tcl config that is placed
    // outside of the virtualenv.  It also sometimes is entirely unable to find
    // the tcl config that comes from the standalone python builds.
    #[cfg(unix)]
    {
        inject_tcl_config(venv, &py_bin)?;
    }

    Ok(())
}

/// Update the cloud synchronization marker for the given path
/// based on the config flag.
fn update_venv_sync_marker(output: CommandOutput, venv_path: &Path) {
    if let Err(err) = mark_path_sync_ignore(venv_path, Config::current().venv_mark_sync_ignore()) {
        if output != CommandOutput::Quiet && Config::current().venv_mark_sync_ignore() {
            warn!(
                "unable to mark virtualenv {} ignored for cloud sync: {}",
                venv_path.display(),
                err
            );
        }
    }
}

#[cfg(unix)]
fn inject_tcl_config(venv: &Path, py_bin: &Path) -> Result<(), Error> {
    let lib_path = match py_bin
        .parent()
        .and_then(|x| x.parent())
        .map(|x| x.join("lib"))
    {
        Some(path) => path,
        None => return Ok(()),
    };

    let mut tcl_lib = None;
    let mut tk_lib = None;

    if let Ok(dir) = lib_path.read_dir() {
        for entry in dir.filter_map(|x| x.ok()) {
            let filename = entry.file_name();
            let name = match filename.to_str() {
                Some(name) => name,
                None => continue,
            };
            if name.starts_with("tcl8") {
                tcl_lib = Some(name.to_string());
                if tk_lib.is_some() {
                    break;
                }
            } else if name.starts_with("tk8") {
                tk_lib = Some(name.to_string());
                if tcl_lib.is_some() {
                    break;
                }
            }
        }
    }

    if tk_lib.is_none() && tcl_lib.is_none() {
        return Ok(());
    }

    if let Some(site_packages) = get_site_packages(venv.join("lib"))? {
        fs::write(
            site_packages.join("_tcl-init.pth"),
            minijinja::render!(
                r#"import os, sys;
                {%- if tcl_lib -%}
                os.environ.setdefault('TCL_LIBRARY', sys.base_prefix + '/lib/{{ tcl_lib }}');
                {%- endif -%}
                {%- if tk_lib -%}
                os.environ.setdefault('TK_LIBRARY', sys.base_prefix + '/lib/{{ tk_lib }}');
                {%- endif -%}"#,
                tcl_lib,
                tk_lib,
            ),
        )?;
    }

    Ok(())
}

// There is only one folder in the venv/lib folder. But in practice, only pypy will use this method in linux
#[cfg(unix)]
fn get_site_packages(lib_dir: PathBuf) -> Result<Option<PathBuf>, Error> {
    let entries = fs::read_dir(lib_dir).context("read venv/lib/ path is fail")?;

    for entry in entries {
        let entry = entry?;
        let metadata = entry.metadata()?;

        if metadata.is_dir() {
            return Ok(Some(entry.path().join("site-packages")));
        }
    }
    Ok(None)
}
